use self::property::Property;
use domo_proto::identifier::Identifier;
use std::collections::HashMap;

pub mod property;

/// Simple Node abstraction.
#[derive(Debug, Clone)]
pub struct Node {
    /// Who owns this node?
    pub parent: Option<Identifier>,
    /// Identifier of node
    identifier: Identifier,
    /// Child nodes
    children: HashMap<Identifier, Node>,
    /// All properties of the node.
    properties: HashMap<String, Property>,
}

impl Node {
    /// Create a new node
    pub fn new(node_id: Identifier) -> Self {
        Node {
            parent: None,
            identifier: node_id,
            children: HashMap::new(),
            properties: HashMap::new(),
        }
    }

    // Return reference to Node's identifier
    pub fn identifier(&self) -> &Identifier {
        &self.identifier
    }

    /// (Over)writes property to self.
    /// Returns a mutable reference to property.
    pub fn insert_property<'a>(&'a mut self, property: Property) -> &'a mut Property {
        self.properties
            .insert(property.name.clone(), property.clone());
        self.properties.get_mut(&property.name).unwrap()
    }

    /// Attempts to remove property from self.
    /// Returns a mutable reference to the added node.
    pub fn remove_property(&mut self, name: String) -> Option<Property> {
        self.properties.remove(&name)
    }

    /// Attempts to add child node to self.
    /// Returns a mutable reference to the added node.
    pub fn add_child<'a>(&'a mut self, mut node: Node) -> Option<&'a mut Node> {
        if self.children.contains_key(&node.identifier) {
            return None;
        }
        node.parent = Some(self.identifier.clone());
        self.children.insert(node.identifier, node.clone());
        self.children.get_mut(&node.identifier)
    }

    /// Attempts to remove child from self.
    /// Returns removed node.
    pub fn remove_child(&mut self, identifier: &Identifier) -> Option<Node> {
        self.children.remove(&identifier).map(|mut n| {
            n.parent = None;
            n
        })
    }

    /// Check if node has an identifier.
    /// Only checks children.
    pub fn has(&self, identifier: &Identifier) -> bool {
        self.children.contains_key(identifier)
    }

    /// Attempts to find given identifier and returns reference to it.
    pub fn find<'a>(&'a self, identifier: &Identifier) -> Option<&'a Node> {
        if self.identifier == *identifier {
            return Some(self);
        }
        for (_, child) in self.children.iter() {
            if let Some(found) = child.find(identifier) {
                return Some(found);
            }
        }
        None
    }

    /// Attempts to find given identifier and returns mutable reference to it.
    pub fn find_mut<'a>(&'a mut self, identifier: &Identifier) -> Option<&'a mut Node> {
        if self.identifier == *identifier {
            return Some(self);
        }
        for (_, child) in self.children.iter_mut() {
            if let Some(found) = child.find_mut(identifier) {
                return Some(found);
            }
        }
        None
    }
}
